package cn.edu.xidian.graduation.modelchecking.model;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import cn.edu.xidian.graduation.modelchecking.util.CollectionUtil;

/**
 * This graph implementation use a map to store graph data.In the map, a key is
 * a state of the Model and the value is states this key(state) can reach to
 * 
 * @author wenqiang.fanwq
 * 
 */
public class SimpleModel implements Model {
	//
	public static final float INIT_STATE_PERCENT = 0.15F;

	private Map<State, Set<State>> graph;
	private int stateSize;
	private long relationSize;

	private Set<State> initStates;

	public SimpleModel() {
		this.graph = new HashMap<State, Set<State>>();
	}

	@Override
	public void addState(State state) {
		checkNullState(state);

		if (containsState(state)) {
			throw new IllegalArgumentException(
					"this state has already exsisted");
		}

		graph.put(state, null);
		++stateSize;
	}

	@Override
	public void addRelation(State beg, State end) {
		checkNullState(beg, end);
		if (!containsState(beg)) {
			addState(beg);
		}
		if (!containsState(end)) {
			addState(end);
		}

		Set<State> successor = getSuccessor(beg);
		if (successor == null) {
			successor = new HashSet<State>();
			graph.put(beg, successor);
		}
		successor.add(end);
		++relationSize;
	}

	@Override
	public Set<State> getSuccessor(State state) {
		checkNullState(state);
		return graph.get(state);
	}

	@Override
	public Set<State> getPrecursor(State state) {
		checkNullState(state);
		Set<State> precusor = null;
		for (State s : graph.keySet()) {
			Set<State> successor = getSuccessor(s);
			if (successor != null && successor.contains(state)) {
				if (precusor == null) {
					precusor = new HashSet<State>();
				}
				precusor.add(s);
			}
		}
		return precusor;
	}

	@Override
	public boolean hasSuccessor(State state) {
		Set<State> successor = getSuccessor(state);
		return successor != null && successor.size() > 0;
	}

	@Override
	public boolean hasPrecursor(State state) {
		Set<State> precusor = getPrecursor(state);
		return precusor != null && precusor.size() > 0;
	}

	@Override
	public boolean hasRelation(State beg, State end) {
		Set<State> successor = getSuccessor(beg);
		if (successor != null) {
			return successor.contains(end);
		}
		return false;
	}

	@Override
	public boolean isEmpty() {
		return stateSize == 0;
	}

	@Override
	public int stateSize() {
		return stateSize;
	}

	@Override
	public long relationSize() {
		return relationSize;
	}

	@Override
	public boolean containsState(State state) {
		return graph.containsKey(state);
	}

	private void checkNullState(State... states) {
		for (State state : states) {
			if (state == null) {
				throw new IllegalArgumentException("state must not be null");
			}
		}
	}

	@Override
	public Set<State> getAllStates() {
		return new HashSet<State>(graph.keySet());
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (State state : getAllStates()) {
			sb.append(state + ": ");
			Set<State> successor = getSuccessor(state);
			if (successor != null) {
				sb.append("------> " + successor);
			}
			sb.append("\n");
		}
		return sb.toString();
	}

	/**
	 * random algo for generating init states which choose 10 percent states of
	 * all states as the init states
	 */
	@Override
	public Set<State> getInitStates() {
		if (initStates == null) {
			Set<State> all = getAllStates();
			List<State> allList = CollectionUtil.list(all);
			int allSize = allList.size();
			if (allSize == 0) {
				return new HashSet(0);
			}

			int initSize = (int) (allSize * INIT_STATE_PERCENT) + 1;
			Set<Integer> randomPositions = new HashSet<Integer>(initSize);
			for (int i = 0; i < initSize; i++) {
				while (true) {
					int position = CollectionUtil.randomInteger(allSize);
					if (!randomPositions.contains(position)) {
						randomPositions.add(position);
						break;
					}
				}
			}
			initStates = new HashSet<State>(initSize);
			for (Integer position : randomPositions) {
				initStates.add(allList.get(position));
			}
		}

		return initStates;
	}

	@Override
	public Set<State> allNoSuccessorStates() {
		Set<State> result = new HashSet<State>();
		Set<State> states = getAllStates();
		for (State state : states) {
			if (!hasSuccessor(state)) {
				result.add(state);
			}
		}
		return result;
	}

	@Override
	public Set<State> allNoPrecusorStaes() {
		Set<State> result = new HashSet<State>();
		Set<State> states = getAllStates();
		for (State state : states) {
			if (!hasPrecursor(state)) {
				result.add(state);
			}
		}
		return result;
	}

}
