package cn.edu.xidian.graduation.modelchecking.util;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import cn.edu.xidian.graduation.modelchecking.model.AbstractState;
import cn.edu.xidian.graduation.modelchecking.model.Conterexample;
import cn.edu.xidian.graduation.modelchecking.model.Model;
import cn.edu.xidian.graduation.modelchecking.model.OriginalState;
import cn.edu.xidian.graduation.modelchecking.model.State;

/**
 * This utility class provides four algorithms for checking failure state in the
 * given couterexample.
 * 
 * @author wenqiang.fanwq
 * 
 */
public class SpuriousChecker {
	// indicate whether the result has been obtained
	private static volatile boolean resultObtained = false;

	public static AbstractState checkSpurious1(Model originalModel,
			Model abstractModel, Conterexample conterexample) {
		List<AbstractState> states = conterexample.getStates();
		for (int i = 0; i < states.size(); i++) {
			AbstractState curState = states.get(i);
			AbstractState preState = null;
			AbstractState postState = null;
			if (i > 0) {
				preState = states.get(i - 1);
			}
			if (i < states.size() - 1) {
				postState = states.get(i + 1);
			}
			if (isFailure(curState, preState, postState, originalModel)) {
				if(i<2*states.size()/3){
					continue;
				}
				return curState;
			}
		}
		return null;
	}

	public static AbstractState checkSpurious2(Model originalModel,
			Model abstractModel, Conterexample conterexample) {
		// sort the states in the conterexample with their weights
		List<Integer> indexes = sortWithWeight(conterexample, originalModel);
		List<AbstractState> states = conterexample.getStates();
		for (int i = 0; i < states.size(); i++) {
			int index = indexes.get(i);
			AbstractState curState = states.get(index);
			AbstractState preState = null;
			AbstractState postState = null;
			if (index > 0) {
				preState = states.get(index - 1);
			}
			if (index < states.size() - 1) {
				postState = states.get(index + 1);
			}
			if (isFailure(curState, preState, postState, originalModel)) {
				return curState;
			}
		}
		return null;
	}

	/**
	 * A parallel algo for checking a given conterexample
	 * 
	 * @param originalModel
	 * @param abstractModel
	 * @param conterexample
	 * @return a spurious abstarct state
	 * @throws InterruptedException
	 */
	public static AbstractState checkSpurious3(final Model originalModel,
			Model abstractModel, Conterexample conterexample)
			throws InterruptedException {
		final ExecutorService executor = Executors.newFixedThreadPool(10);
		final CompletionService<AbstractState> completionService = new ExecutorCompletionService<AbstractState>(
				executor);

		final List<AbstractState> states = conterexample.getStates();
		resultObtained = false;

		executor.execute(new Runnable() {

			@Override
			public void run() {
				for (int i = 0; i < states.size() && !resultObtained; i++) {
					AbstractState curState = states.get(i);
					AbstractState preState = null;
					AbstractState postState = null;
					if (i > 0) {
						preState = states.get(i - 1);
					}
					if (i < states.size() - 1) {
						postState = states.get(i + 1);
					}
					completionService.submit(new Task(curState, preState,
							postState, originalModel));

				}
			}
		});

		try {
			for (int i = 0; i < states.size(); i++) {
				Future<AbstractState> future = completionService.take();
				try {
					if (future.get() != null) {
						resultObtained = true;
						return future.get();
					}
				} catch (ExecutionException e) {
					e.printStackTrace();
				}
			}
		} finally {
			executor.shutdownNow();
		}

		return null;
	}

	public static AbstractState checkSpurious4(final Model originalModel,
			Model abstractModel, Conterexample conterexample)
			throws InterruptedException {
		ExecutorService executor = Executors.newFixedThreadPool(10);
		final CompletionService<AbstractState> completionService = new ExecutorCompletionService<AbstractState>(
				executor);

		final List<Integer> indexs = sortWithWeight(conterexample,
				originalModel);
		final List<AbstractState> states = conterexample.getStates();
		resultObtained = false;
		executor.execute(new Runnable() {

			@Override
			public void run() {
				for (int i = 0; i < states.size() && !resultObtained; i++) {
					int index = indexs.get(i);
					AbstractState curState = states.get(index);
					AbstractState preState = null;
					AbstractState postState = null;
					if (index > 0) {
						preState = states.get(index - 1);
					}
					if (index < states.size() - 1) {
						postState = states.get(index + 1);
					}
					completionService.submit(new Task(curState, preState,
							postState, originalModel));

				}

			}

		});

		try {
			for (int i = 0; i < states.size(); i++) {
				Future<AbstractState> future = completionService.take();
				try {
					if (future.get() != null) {
						resultObtained = true;
						return future.get();
					}
				} catch (ExecutionException e) {
					e.printStackTrace();
				}
			}
		} finally {
			executor.shutdownNow();
		}

		return null;
	}

	/**
	 * SPLITPATH algo
	 */
	public static AbstractState checkSpuriousWithSplithPath(
			Model originalModel, Model abstractModel,
			Conterexample conterexample) {
		List<AbstractState> states = conterexample.getStates();
		Set<State> m = CollectionUtil.intersection(originalModel
				.getInitStates(), transform(states.get(0).getData()));
		if (m== null) {
			return states.get(0);
		}
		for (int i = 1; i < states.size(); i++) {
			m = fallInto(m, states.get(i).getData(), originalModel);
			if (m == null || m.size()==0) {
				if(i<=2*states.size()/3){
					continue;
				}
				return states.get(i - 1);
			}
		}
		return null;
	}

	/**
	 * A task to check whether or not the given state is spurious.
	 * 
	 * @author wenqiang.fanwq
	 * 
	 */
	private static class Task implements Callable<AbstractState> {
		private AbstractState curState;
		private AbstractState preState;
		private AbstractState postState;
		private Model originalModel;

		public Task(AbstractState curState, AbstractState preState,
				AbstractState postState, Model originalModel) {
			super();
			this.curState = curState;
			this.preState = preState;
			this.postState = postState;
			this.originalModel = originalModel;
		}

		@Override
		public AbstractState call() throws Exception {
			if (isFailure(curState, preState, postState, originalModel)) {
				return curState;
			}
			return null;
		}

	}

	/**
	 * This method provide an algorithm for obtain the set of states this
	 * abstarct state that are reachable from some state in the previous state
	 * 
	 * @param preState
	 * @return the set of original states in this abstract state that can be
	 *         reached from the previous state in a given conterexample
	 */
	public static Set<State> canBeReachedFrom(AbstractState curState,
			AbstractState preState, Model originalModel) {
		Set<State> result = new HashSet<State>();
		Set<State> pre = null;

		if (preState == null) {// curState is the first state in the
								// conterexample
			// Set<State> noPrecursorStates =
			// originalModel.allNoPrecusorStaes();
			Set<State> data = transform(curState.getData());
			pre = CollectionUtil.intersection(data,
					originalModel.getInitStates());
		} else {
			pre = in(preState.getData(), curState.getData(), originalModel);
		}
		while (true) {
			result = CollectionUtil.union(result, pre);
			Set<State> newPre = in(pre, curState.getData(), originalModel);
			if (newPre.equals(pre)) {
				break;
			}
			pre = newPre;
		}
		result = CollectionUtil.union(result, pre);

		return result;
	}

	/**
	 * @param postState
	 * @return the set of states in this abstract state from which some states
	 *         in the next state are reachable
	 */
	public static Set<State> canReachTo(AbstractState curState,
			AbstractState postState, Model originalModel) {
		Set<State> result = new HashSet<State>();
		Set<State> post = null;
		if (postState == null) { // curState is the last state in the given
									// conterexample
			Set<State> noSuccessorStates = originalModel.allNoSuccessorStates();
			Set<State> data = transform(curState.getData());
			post = CollectionUtil.intersection(data, noSuccessorStates);

		} else {
			post = out(curState.getData(), postState.getData(), originalModel);
		}
		while (true) {
			result = CollectionUtil.union(result, post);
			Set<State> newPost = out(curState.getData(), post, originalModel);
			if (newPost.equals(post)) {
				break;
			}
			post = newPost;
		}
		result = CollectionUtil.union(result, post);

		return result;
	}

	private static Set<State> fallInto(Set<? extends State> from,
			Set<? extends State> to, Model originalModel) {
		Set<State> result = new HashSet<State>();
		Set<State> reachable = in(from, to, originalModel);
		if (reachable == null) {
			return null;
		}
		while (true) {
			result = CollectionUtil.union(result, reachable);
			Set<State> newReachable = in(reachable, to, originalModel);
			if (reachable.equals(newReachable)) {
				break;
			}
			reachable = newReachable;
		}
		result = CollectionUtil.union(result, reachable);
		return result;
	}

	/**
	 * For each state in cur if a state in pre has a edge to it, put this state
	 * in the result set
	 */
	private static Set<State> in(Set<? extends State> pre,
			Set<? extends State> cur, Model originalModel) {
		if (pre == null || cur == null) {
			return null;
		}
		Set<State> result = new HashSet<State>();

		Iterator<? extends State> curIter = cur.iterator();
		Iterator<? extends State> preIter = pre.iterator();
		while (curIter.hasNext()) {
			State curState = curIter.next();
			while (preIter.hasNext()) {
				State preState = preIter.next();
				if (originalModel.hasRelation(preState, curState)) {
					result.add(curState);
					break;
				}
			}
		}
		return result;
	}

	/**
	 * For each state in cur that can reach to the state in post, put this state
	 * in the result set
	 */
	private static Set<State> out(Set<? extends State> cur,
			Set<? extends State> post, Model originalModel) {
		if (post == null || cur == null) {
			return null;
		}
		Set<State> result = new HashSet<State>();

		Iterator<? extends State> curIter = cur.iterator();
		Iterator<? extends State> postIter = post.iterator();
		while (curIter.hasNext()) {
			State curState = curIter.next();
			while (postIter.hasNext()) {
				State postState = postIter.next();
				if (originalModel.hasRelation(curState, postState)) {
					result.add(curState);
					break;
				}
			}
		}
		return result;
	}

	public static boolean isFailure(AbstractState curState,
			AbstractState preState, AbstractState postState, Model originalModel) {
		Set<State> in = canBeReachedFrom(curState, preState, originalModel);
		Set<State> out = canReachTo(curState, postState, originalModel);

		Set<State> intersection = CollectionUtil.intersection(in, out);
		if (intersection == null || intersection.size() == 0) {
			return true;
		}
		return false;
	}

	/**
	 * @return
	 */
	public static long getWeight(AbstractState state, Model originalModel) {
		return (long) ein(state, originalModel) * eout(state, originalModel);
	}

	private static int ein(AbstractState abstractState, Model originalModel) {
		Set<OriginalState> states = abstractState.getData();
		int result = 0;
		for (OriginalState state : states) {
			Set<State> precusor = originalModel.getPrecursor(state);
			if (precusor != null) {
				precusor.removeAll(states);
				result += precusor.size();
			}
		}
		return result;
	}

	private static int eout(AbstractState abstractState, Model originalModel) {
		Set<OriginalState> states = abstractState.getData();
		int result = 0;
		for (OriginalState state : states) {
			Set<State> successor = originalModel.getSuccessor(state);
			if (successor != null) {
				successor.removeAll(states);
				result += successor.size();
			}
		}
		return result;
	}

	private static List<Integer> sortWithWeight(Conterexample conterexample,
			Model originalModel) {
		List<AbstractState> states = conterexample.getStates();
		Map<Integer, Long> weights = new HashMap<Integer, Long>();
		for (int i = 0; i < states.size(); i++) {
			weights.put(i, states.get(i).getWeight());
		}

		List<Integer> w = CollectionUtil.list(weights.keySet());
		Collections.sort(w, new WeightComparator(weights));
		return w;
	}

	private static class WeightComparator implements Comparator<Integer> {
		private Map<Integer, Long> weights;

		public WeightComparator(Map<Integer, Long> weights) {
			this.weights = weights;
		}

		@Override
		public int compare(Integer o1, Integer o2) {
			long w1 = weights.get(o1);
			long w2 = weights.get(o2);
			if (w1 < w2) {
				return 1;
			}
			if (w1 > w2) {
				return -1;
			}
			return 0;
		}

	}

	// util
	public static Set<State> transform(Set<OriginalState> states) {
		Set<State> result = new HashSet<State>();
		for (OriginalState s : states) {
			result.add(s);
		}
		return result;
	}
}
